from .mads import Gads

from .encoder import GEncoder

from .mads import Network
from .archgen import AG

from .genotypes import NA_PRIMITIVES, LA_PRIMITIVES, POOL_PRIMITIVES, READOUT_PRIMITIVES, ACT_PRIMITIVES